from __future__ import print_function
from __future__ import unicode_literals

from functools import wraps

import boto3
import boto3_session_cache
import botocore.session
from botocore.exceptions import ClientError, ParamValidationError

from . import config


_sessions = {}


def retry(attempts=3):
    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            tries = attempts
            while True:
                tries -= 1
                try:
                    return func(*args, **kwargs)
                except (ClientError, ParamValidationError) as error:
                    if tries > 0:
                        print('[ssha] {}'.format(error))
                    else:
                        raise
        return wrapped
    return wrapper


@retry()
def client(*args, **kwargs):
    return session().client(*args, **kwargs)


@retry()
def credentials(*args, **kwargs):
    return session().get_credentials(*args, **kwargs)


@retry()
def resource(*args, **kwargs):
    return session().resource(*args, **kwargs)


def session():
    aws_config = config.get('aws') or {}
    profile = aws_config.pop('profile_name', None)
    if 'botocore' not in _sessions:
        print('[ssha] creating aws session')
        _sessions['botocore'] = botocore.session.Session(profile=profile)
        resolver = _sessions['botocore'].get_component('credential_provider')
        provider = resolver.get_provider('assume-role')
        provider.cache = boto3_session_cache.JSONFileCache()
    return boto3.Session(botocore_session=_sessions['botocore'], **aws_config)



from __future__ import print_function

import argparse

from . import __version__, config, ec2, menu, settings, ssm, ssh


def main():
    try:

        parser = argparse.ArgumentParser(prog=__package__)
        parser.add_argument('config', nargs='?', help='Configuration name')
        parser.add_argument('search', nargs='?', help='Instance search string')
        parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')
        parser.add_argument('--region', help='Region name')
        parser.add_argument('--settings', help='Path to the .ssha file')
        parser.add_argument('--version', action='store_true', help='show the installed version and exit')

        args = parser.parse_args()

        if args.version:
            print(__version__)
            return 0

        settings.load(settings_path=args.settings, verbose=args.verbose)

        configs = config.names()
        config_name = menu.choose_config(configs, args.config)
        config.load(config_name)

        regions = config.regions()
        region_name = menu.choose_config(regions, args.region)
        if region_name:
            config.add('aws.region_name', region_name)

        instances = ec2.discover_instances()
        instance = menu.choose_instance(instances, args.search)
        if instance:

            if config.get('bastion') and not config.get('bastion.disabled'):
                bastion = ec2.discover_bastion(instance)
            else:
                bastion = None

            if config.get('ssm'):
                ssm.send_command(instance, bastion)

            ssh.connect(instance, bastion)

        else:
            print('[ssha] no matching instances found')

    except KeyboardInterrupt:
        pass



from __future__ import unicode_literals

import copy
import os
import re
import subprocess
import tempfile

from fnmatch import fnmatch
from paramiko.config import SSHConfig

from . import errors, settings

try:
    basestring
except NameError:
    basestring = str


_config = {}
_ssh_config = {}
_tempfiles = {}


def _exec(command):
    return subprocess.check_output(command, shell=True).strip().decode('utf-8')


def _exec_all(data):
    if isinstance(data, basestring):
        for var in re.findall('\$\{(.+?)\}', data):
            data = data.replace('${' + var + '}', get(var))
        if data.startswith('$(') and data.endswith(')'):
            data = _exec(data[2:-1])
    elif isinstance(data, dict):
        for key, value in data.items():
            data[key] = _exec_all(value)
    elif isinstance(data, list):
        data = [_exec_all(item) for item in data]
    return data


def _get(key, default=None):
    value = _config
    for key in key.split('.'):
        value = value.get(key)
        if not value:
            break
    return value or default


def _get_ssh_config(key):
    if not _ssh_config:

        ssh_config = SSHConfig()

        path = os.path.expanduser('~/.ssh/config')
        if os.path.exists(path):
            with open(path) as open_file:
                ssh_config.parse(open_file)

        # Create a fake hostname like "dev.myproject.ssha" to allow users to
        # set options in ~/.ssh/config based on the environment and project.
        hostname_parts = (get('config.name'), get('ssha.name'), 'ssha')
        hostname = '.'.join(filter(None, hostname_parts))

        result = ssh_config.lookup(hostname)

        if 'identityfile' not in result:
            result['identityfile'] = [
                '~/.ssh/id_rsa',
                '~/.ssh/id_dsa',
                '~/.ssh/id_ecdsa',
                '~/.ssh/id_ed25519',
            ]

        if 'user' not in result:
            user = os.environ.get('USER')
            if user:
                result['user'] = user

        _ssh_config.update(result)

    return _ssh_config.get(key, [])


def _is_used(data, key):
    """
    Checks if a key is used as a variable anywhere else in the config.

    """
    if isinstance(data, basestring):
        for var in re.findall('\$\{(.+?)\}', data):
            if var == key:
                return True
    elif isinstance(data, dict):
        for value in data.values():
            if _is_used(value, key):
                return True
    elif isinstance(data, list):
        for value in data:
            if _is_used(value, key):
                return True
    return False


def _merge(target, source):
    for key in source:
        if key in target:
            if isinstance(target[key], dict) and isinstance(source[key], dict):
                _merge(target[key], source[key])
            else:
                target[key] = source[key]
        else:
            target[key] = source[key]


def add(name, value):
    data = {}
    here = data
    for key in name.split('.'):
        last = here
        here[key] = {}
        here = here[key]
    last[key] = value
    update(data)


def get(key, default=None):

    value = _get(key, default)

    if not value:
        return value

    if isinstance(value, (dict, list)):
        value = copy.deepcopy(value)

    return _exec_all(value)


def load(name):
    """
    Loads a config from the settings.

    """

    update(settings.all())

    config_specific_settings = _config.pop('config', None) or {}
    if name:
        if name not in names():
            errors.string_exit('config {} not found in .ssha file'.format(name))
        for config_name in config_specific_settings:
            if fnmatch(name, config_name):
                update(config_specific_settings[config_name])
        add('config.name', name)

    iam_group_specific_settings = get('iam.group')
    if iam_group_specific_settings:
        from . import iam
        for group in iam.groups():
            if group in iam_group_specific_settings:
                update(iam_group_specific_settings[group])

    # Default to SSH's default user.
    if not _get('ssh.username'):
        user = _get_ssh_config('user')
        if user:
            add('ssh.username', user)

    if _get('bastion') and not _get('ssh.proxy_command'):
        from . import ssh
        if _get('ssh.user_known_hosts_file'):
            known_hosts_options = ['-o', 'UserKnownHostsFile=${ssh.user_known_hosts_file}']
        else:
            known_hosts_options = []
        proxy_command = ['ssh'] + known_hosts_options + ['-W', '%h:%p', '${bastion.address}']
        add('ssh.proxy_command', ssh.format_command(proxy_command))

    # To support configs like this:
    #   ssm.parameters.key = ["$(cat '${ssh.identityfile_public}')"]
    # If "ssh.identityfile_public" has been used as a variable, then find the
    # first identity file in the SSH config that has a matching ".pub" file.
    # The SSM document command would then add this public key to a user account
    # so that SSH key-based authentication will work.
    if is_used_as_variable('ssh.identityfile_public') and not _get('ssh.identityfile_public'):
        for private_key_path in _get_ssh_config('identityfile'):
            private_key_path = os.path.expanduser(private_key_path)
            if os.path.exists(private_key_path):
                public_key_path = private_key_path + '.pub'
                if os.path.exists(public_key_path):
                    add('ssh.identityfile_public', public_key_path)
                    break

    # To support configs like this:
    #   ssh.user_known_hosts_file = "${ssm.host_keys_file}"
    # If "ssm.host_keys_file" has been used as a variable, then create a temp
    # file and set "ssm.host_keys_file" as the temp file path. The SSM code
    # will populate the file with the output of the SSM command. The SSM
    # command must print the server's SSH host keys for this to work.
    if is_used_as_variable('ssm.host_keys_file') and not _get('ssm.host_keys_file'):
        _tempfiles['host_keys_file'] = tempfile.NamedTemporaryFile(suffix='-ssha-known-hosts')
        add('ssm.host_keys_file', _tempfiles['host_keys_file'].name)


def names():
    ssha_settings = settings.all().get('ssha') or {}
    return ssha_settings.get('configs') or []


def regions():
    region = get('aws.region_name')
    if region:
        return [region]
    return get('ssha.regions') or []


def reset():
    _config.clear()
    _ssh_config.clear()
    _tempfiles.clear()


def update(data):
    _merge(_config, data)


def is_used_as_variable(key):
    return _is_used(_config, key)



from __future__ import print_function
from __future__ import unicode_literals

import copy
import hcl
import operator
import os
import re

from distutils.version import StrictVersion

from . import errors, __version__


_settings = {}

operators = {
    None: operator.eq,
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}

pattern = re.compile(
    r'\s*'
    r'(?P<operator>[<=>!]{1,2})?\s*'
    r'(?P<version>.+)\s*'
)


def _find_settings_path(path):
    if path:
        if os.path.isfile(path):
            return path
        errors.string_exit('Could not find .ssha file: {}'.format(path))

    cwd = os.path.realpath(os.getcwd())

    while cwd and cwd != '/':
        path = os.path.join(cwd, '.ssha')
        if os.path.isfile(path):
            return path
        cwd = os.path.dirname(cwd)

    errors.string_exit('Could not find .ssha file in current directory or parent directories')


def _load(path):
    try:
        with open(path) as settings_file:
            return hcl.load(settings_file)
    except IOError as error:
        errors.string_exit('Error reading settings: {}'.format(error))
    except Exception as error:
        errors.string_exit('Error parsing settings: {}'.format(error))


def _validate_version(data):
    ssha_settings = data.get('ssha') or {}
    version = ssha_settings.get('version')
    if version is None:
        return

    ssha_version = StrictVersion(__version__)
    requirements = version.split(',')
    for requirement in requirements:
        matched = pattern.match(requirement)
        if matched is None:
            errors.string_exit('Error parsing ssha version: {}'.format(requirement))

        spec_operator = matched.group('operator')
        if spec_operator not in operators:
            errors.string_exit('Error parsing operator in ssha version: {}'.format(requirement))

        try:
            spec_version = StrictVersion(matched.group('version'))
        except ValueError as error:
            errors.string_exit('Error parsing ssha version: {}'.format(error))

        if not operators[spec_operator](ssha_version, spec_version):
            errors.string_exit("Current ssha version {} doesn't meet requirements in settings: {}".format(
                ssha_version, requirement
            ))


def all():
    return copy.deepcopy(_settings)


def load(**defaults):
    if defaults.get('verbose'):
        print('[ssha] finding settings file')
    update(defaults)

    settings_path = defaults.get('settings_path')
    path = _find_settings_path(settings_path)

    if defaults.get('verbose'):
        print('[ssha] loading {}'.format(path))
    data = _load(path)
    _validate_version(data)
    update(data)


def reset():
    _settings.clear()


def update(data):
    _settings.update(data)



from __future__ import print_function

import operator

from . import aws, config, errors, ssm


_instances = []


def _describe_instances():

    ec2 = aws.client('ec2')

    print('[ssha] discovering ec2 instances')

    paginator = ec2.get_paginator('describe_instances')
    page_iterator = paginator.paginate(MaxResults=15)
    for page in page_iterator:
        if page['ResponseMetadata']['HTTPStatusCode'] != 200:
            errors.json_exit(page)
        for reservation in page['Reservations']:
            for instance in reservation['Instances']:
                yield instance


def _filter_instances(instances, rules):
    result = []
    for instance in instances:
        if _rules_pass(instance, rules):
            result.append(instance)
    return result


def _instance_sort_key(instance):
    result = []
    for field in config.get('display.sort') or ['InstanceId']:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
            if not value:
                break
        result.append(value or '')
    return result


def _rules_pass(obj, rules, compare=operator.eq):

    for key, expected_value in rules.items():

        if isinstance(expected_value, dict):

            if key.endswith('NotEqual'):
                nested_compare = operator.ne
                key = key[:-len('NotEqual')]
            else:
                nested_compare = compare

            nested_rules_passed = _rules_pass(
                obj=obj.get(key) or {},
                rules=expected_value,
                compare=nested_compare,
            )
            if not nested_rules_passed:
                return False

        elif not compare(obj.get(key), expected_value):
            return False

    return True


def _find_instances(ec2_filters, ssm_filters):

    if not _instances:
        for instance in _describe_instances():
            tags = {}
            for tag in instance.get('Tags') or []:
                tags[tag['Key']] = tag['Value']
            instance['Tags'] = tags
            _instances.append(instance)

    instances = _instances

    if instances and ec2_filters:
        instances = _filter_instances(instances, ec2_filters)

    if instances and ssm_filters:
        for info in ssm.find_instances():
            for instance in instances:
                if instance['InstanceId'] == info['InstanceId']:
                    instance.update(info)
                    break
        instances = _filter_instances(instances, ssm_filters)

    return instances


def discover_bastion(instance):
    bastions = _find_instances(
        ec2_filters=config.get('bastion.ec2'),
        ssm_filters=config.get('bastion.ssm'),
    )
    for bastion in bastions:
        if bastion['InstanceId'] == instance['InstanceId']:
            return None
    if bastions:
        return bastions[0]
    else:
        errors.string_exit('Bastion not found')


def discover_instances():
    instances = _find_instances(
        ec2_filters=config.get('discover.ec2'),
        ssm_filters=config.get('discover.ssm'),
    )
    return sorted(instances, key=_instance_sort_key)


def label(instance):
    result = []
    for field in config.get('display.fields') or []:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
        result.append(value or '')
    return result or [instance['InstanceId']]



from __future__ import print_function
from __future__ import unicode_literals

import re
import time

from botocore.exceptions import ClientError

from . import aws, config, errors, ssh


def _send_command(instance_ids):

    document_name = config.get('ssm.document.name')
    parameters = config.get('ssm.parameters')

    print('[ssha] ssm send {document} to {instances}'.format(
        document=document_name,
        instances=' and '.join(instance_ids),
    ))

    ssm = aws.client('ssm')

    result = ssm.send_command(
        InstanceIds=instance_ids,
        DocumentName=document_name,
        Parameters=parameters,
    )

    try:
        command_id = result['Command']['CommandId']
    except Exception:
        errors.json_exit(result)

    return command_id


def _wait_for_command(instance_ids, command_id):

    ssm = aws.client('ssm')

    outputs = {}

    for instance_id in instance_ids:

        result = {'Status': 'Pending'}

        while result['Status'] in ('Pending', 'InProgress', 'Delayed'):
            time.sleep(0.25)
            try:
                result = ssm.get_command_invocation(
                    CommandId=command_id,
                    InstanceId=instance_id,
                )
            except ClientError:
                pass

        if result['Status'] != 'Success':
            errors.json_exit(result)

        print('[ssha] ssm command finished on {}'.format(instance_id))
        outputs[instance_id] = result['StandardOutputContent']

    return outputs


def find_instances():

    ssm = aws.client('ssm')
    paginator = ssm.get_paginator('describe_instance_information')
    page_iterator = paginator.paginate()

    instance_info_list = []
    for page in page_iterator:
        if page['ResponseMetadata']['HTTPStatusCode'] != 200:
            errors.json_exit(page)

        instance_info_list += page['InstanceInformationList']

    return instance_info_list


def send_command(instance, bastion):

    instance_ids = [inst['InstanceId'] for inst in (instance, bastion) if inst]

    command_id = _send_command(
        instance_ids=instance_ids,
    )

    outputs = _wait_for_command(
        instance_ids=instance_ids,
        command_id=command_id,
    )

    host_keys_file = config.get('ssm.host_keys_file')
    if host_keys_file:
        instance_ips = {}

        if bastion:
            instance_id = bastion['InstanceId']
            instance_ips[instance_id] = ssh.get_ip(bastion, connect_through_bastion=False)

        instance_id = instance['InstanceId']
        instance_ips[instance_id] = ssh.get_ip(instance, connect_through_bastion=bool(bastion))

        with open(host_keys_file, 'w') as open_file:
            for instance_id, ip in instance_ips.items():
                host_keys = outputs[instance_id]
                for line in host_keys.splitlines():
                    # Replace hostname from the host keys line
                    # with the instance's hostname.
                    line = re.sub(r'^[^#\s]+', ip, line)
                    open_file.write(line + '\n')
